;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; led 20180101 - a simple line-oriented text editor ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Copyright (C) 2017 Pasquale Frega
; All rights reserved

; Released under the semplified BSD license

; last edit: 06-Jan-2018

; begin

; Scroll _list from _start by _step, typing s stop this, return _start _step and _line_number

(defun scroll (_list _start _step _line_number)
 (set! _list (drop _list (- _start 1)))
 (with (_got_input nil)
  (if (not _line_number)
   (for-each _list (groups-of _list _step)
    (when (not (eq? _got_input "s"))
     (for-each _string _list
      (print! _string)
      (inc! _start)
     )
     (set! _got_input (io/read))
    )
   )
   (for-each _list (groups-of _list _step)
    (when (not (eq? _got_input "s"))
     (for-each _string _list
      (printf "%d\t%s" _start _string)
      (inc! _start)
     )
     (set! _got_input (io/read))
    )
   )
  )
 )
 (list _start _step _line_number)
)

; A file name is mandatory

(when (= (n arg) 0)
 (print! "Please specify a file name.")
 (exit! 3)
)

(import io (read-lines! write-lines!))
(import lua/os (execute))
(let* [(_file_name (car arg)) (_buffer (read-lines! _file_name)) (_got_list '()) (_clipboard '()) (_from 0) (_to 0) (_to_insert nil) (_record_scroll '(1 40)) (_previous_buffer _buffer) (_shell_command "")]
 (if (not _buffer)
  (progn
   (print! "File open failed, i will try to create a new one.")
   (set! _buffer '())
  )
  (print! "File open successful.")
 )
 
 ; Quit unconditionally
 
 (while (not (eq? (car _got_list) "q"))

  (io/write "led>")
  (io/flush)
  (when-with (_got_input (io/read))
   (set! _got_list (string/split _got_input ","))
   
   ; Check for any comma
 
   (for _index 1 (n _got_list) 1
    (when (string/match (nth _got_list _index) "|cM")
     (setq! (nth _got_list _index) (string/gsub (nth _got_list _index) "|cM" ","))
    )
   )
 
   (cond 
    
    ; Copy all or some or only one line
 
    ((eq? (car _got_list) "c")
     (set! _clipboard '())
     (cond
      ((= (n _got_list) 1)
       (set! _clipboard _buffer)
      )
      ((and (= (n _got_list) 2) (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (set! _clipboard (list (nth _buffer (string->number (cadr _got_list)))))
      )
      ((and (= (n _got_list) 3) (string->number (cadr _got_list)) (string->number (caddr _got_list)))
       (when (and (nth _buffer (string->number (cadr _got_list))) (nth _buffer (string->number (caddr _got_list))))
        (for _line (string->number (cadr _got_list)) (string->number (caddr _got_list)) 1
         (set! _clipboard (snoc _clipboard (nth _buffer _line)))
        )
       )
      )
      (true)
     )
    )
    
    ; Delete all or some or only one line
    
    ((eq? (car _got_list) "d")
     (cond
      ((= (n _got_list) 1)
       (set! _previous_buffer _buffer)
       (set! _buffer '())
      )
      ((and (= (n _got_list) 2) (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (set! _previous_buffer (snoc _buffer))
       (remove-nth! _buffer (string->number (cadr _got_list)))
      )
      ((and (= (n _got_list) 3) (string->number (cadr _got_list)) (string->number (caddr _got_list)))
       (when (and (nth _buffer (string->number (cadr _got_list))) (nth _buffer (string->number (caddr _got_list))))
        (set! _previous_buffer (snoc _buffer))
        (for _line (string->number (cadr _got_list)) (string->number (caddr _got_list)) 1 
         (remove-nth! _buffer (string->number (cadr _got_list)))
        )
       )
      )
      (true)
     )      
    )
    
    ; Edit a line (only on Amiga's get it from clipboard with RAmiga+V), with i start to insert from that line, type a dot to exit
    
    ((eq? (car _got_list) "e")
     (when (and (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
      (set! _previous_buffer (snoc _buffer))
      (with (_chid (io/open "C:Clip" "rb"))
       (if (not _chid)
        (print! (nth _buffer (string->number (cadr _got_list))))
        (let [(_old_string (nth _buffer (string->number (cadr _got_list)))) (_new_string "")]
         (io/close _chid)
         (for-each _char (string/string->chars _old_string)
          (if (string/match _char "%a")
           (set! _new_string (sprintf "%s%s" _new_string _char))
           (set! _new_string (sprintf "%s*%s" _new_string _char))
          )
         )
         (execute (sprintf "C:Run >NIL: C:Clip set \"%s\"" _new_string))
        )
       )
       (if (= (n _got_list) 2)
        (setq! (nth _buffer (string->number (cadr _got_list))) (io/read))
        (progn
         (remove-nth! _buffer (string->number (cadr _got_list)))
         (set! _to_insert '())
         (while (not (eq? (last _to_insert) "."))
          (set! _to_insert (snoc _to_insert (io/read)))
         )
         (pop-last! _to_insert)
         (for-each _string (reverse _to_insert)
          (insert-nth! _buffer (string->number (cadr _got_list)) _string)
         )
        )
       )
      )
     )
    )
    
    ; Show file infos or change file name
    
    ((eq? (car _got_list) "f")
     (if (= (n _got_list) 1)
      (if (empty? _buffer)
       (printf "Name: %s  Lines: 0  Bytes: 0  Format:" _file_name)
       (let [(_bytes (n _buffer)) (_format "Unix")]
        (for-each _string _buffer
         (set! _bytes (+ _bytes (string/len _string)))
        )
        (dec! _bytes)
        (when (string/ends-with? (car _buffer) "\r")
         (set! _format "DOS/Mac")
        )
        (printf "Name: %s  Lines: %d  Bytes: %d  Format: %s" _file_name (n _buffer) _bytes _format)
       )
      )
      (set! _file_name (string/trim (cadr _got_list)))
     )
    )
    
    ; Mark a range of lines, list that range with line number
 
    ((eq? (car _got_list) "g")
     (if (= (n _got_list) 1)
      (for _line _from _to 1
       (pcall printf "%d\t%s" _line (nth _buffer _line))
      )
      (when (and (string->number (cadr _got_list)) (string->number (caddr _got_list)) (> (string->number (caddr _got_list)) (string->number (cadr _got_list))))
       (when (and (nth _buffer (string->number (cadr _got_list))) (nth _buffer (string->number (caddr _got_list))))
        (set! _from (string->number (cadr _got_list)))
        (set! _to (string->number (caddr _got_list)))
       )
      )
     )
    )
 
    ; Show help
 
    ((eq? (car _got_list) "h")
     (print! "Usage: lua led.lua FILE\n")
     (print! "Commands are:")
     (print! "  c[,#,#]     Copy all or some or only one line")
     (print! "  d[,#,#]     Delete all or some or only one line")
     (print! "  e,#[,i]     Edit a line (only on Amiga's get it from clipboard with RAmiga+V), with i start to insert from that line, type a dot to exit")
     (print! "  f[,$]       Show file infos or change file name")
     (print! "  g[,#,#]     Mark a range of lines, list that range with line number")
     (print! "  h           Show this help")
     (print! "  i[,#]       Insert into buffer given text from last or from any line, type a dot to exit")
     (print! "  j,#,#       Join a line with another")
     (print! "  k           List marked range without line number")
     (print! "  l[,#,#]     List all or some or only one line, without line number")
     (print! "  m[,#,n]     Press Return to scroll # lines every time (default 40), with n show line number, type s to stop, 'm,' to continue")
     (print! "  n[,#,#]     List all or some or only one line, with line number")
     (print! "  o           Switch to DOS/Mac format")
     (print! "  p[,#]       Paste content of clipboard from last or from any line")
     (print! "  q           Quit unconditionally")
     (print! "  r,$         Read a file and append content into buffer")
     (print! "  s,$[,$,#,#] Search a string or substitute it with another one (support Lua patterns) in all or some or only one line, for a comma type |cM")
     (print! "  t[,#]       Replace tab with # spaces (default 1, usefull on Amiga's)")
     (print! "  u           Undo<->Redo")
     (print! "  v           Show version and license")
     (print! "  w           Write out unconditionally")
     (print! "  x[,$,r]     Execute file or a shell command + file, execute a shell command, with r record it")
    )
 
    ; Insert into buffer given text from last or from any line, type a dot to exit
    
    ((eq? (car _got_list) "i")
     (set! _to_insert '())
     (if (= (n _got_list) 1)
      (pcall printf "%s" (last _buffer))
      (when (and (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (pcall printf "%s" (nth _buffer (- (string->number (cadr _got_list)) 1)))
      )
     )
     (while (not (eq? (last _to_insert) "."))
      (set! _to_insert (snoc _to_insert (io/read)))
     )
     (pop-last! _to_insert)
     (if (= (n _got_list) 1)
      (progn
       (set! _previous_buffer _buffer)
       (set! _buffer (append _buffer _to_insert))
      )
      (when (and (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (set! _previous_buffer (snoc _buffer))
       (for-each _string (reverse _to_insert)
        (insert-nth! _buffer (string->number (cadr _got_list)) _string)
       )
      )
     )
    )
    
    ; Join a line with another
 
    ((eq? (car _got_list) "j")
     (when (and (= (n _got_list) 3) (string->number (cadr _got_list)) (string->number (caddr _got_list)))
      (when (and (nth _buffer (string->number (cadr _got_list))) (nth _buffer (string->number (caddr _got_list))))
       (set! _previous_buffer (snoc _buffer))
       (setq! (nth _buffer (string->number (cadr _got_list))) (sprintf "%s%s" (nth _buffer (string->number (cadr _got_list))) (nth _buffer (string->number (caddr _got_list)))))
      )
     )
    )
 
    ; List marked range without line number

    ((eq? (car _got_list) "k")
     (for _line _from _to 1
      (pcall printf "%s" (nth _buffer _line))
     )
    )

    ; List all or some or only one line, without line number
    
    ((eq? (car _got_list) "l")
     (cond
      ((= (n _got_list) 1)
       (for _line 1 (n _buffer) 1
        (print! (nth _buffer _line))
       )
      )
      ((and (= (n _got_list) 2) (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (print! (nth _buffer (string->number (cadr _got_list))))
      )
      ((and (= (n _got_list) 3) (string->number (cadr _got_list)) (string->number (caddr _got_list)))
       (when (and (nth _buffer (string->number (cadr _got_list))) (nth _buffer (string->number (caddr _got_list))))
        (for _line (string->number (cadr _got_list)) (string->number (caddr _got_list)) 1 
         (print! (nth _buffer _line))
        )
       )
      )
      (true)
     )      
    )
    
    ; Press Return to scroll # lines every time (default 40), with n show line number, type s to stop, 'm,' to continue
 
    ((eq? (car _got_list) "m")
     (cond
      ((= (n _got_list) 1)
       (set! _record_scroll (scroll _buffer 1 40 false)) 
      )
      ((and (= (n _got_list) 2) (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (set! _record_scroll (scroll _buffer 1 (string->number (cadr _got_list)) false)) 
      )
      ((and (= (n _got_list) 3) (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (set! _record_scroll (scroll _buffer 1 (string->number (cadr _got_list)) true)) 
      )
      (true
       (set! _record_scroll (scroll _buffer (car _record_scroll) (cadr _record_scroll) (caddr _record_scroll)))
      )
     )
    )
 
    ; List all or some or only one line, with line number
    
    ((eq? (car _got_list) "n")
     (cond
      ((= (n _got_list) 1)
       (for _line 1 (n _buffer) 1
        (printf "%d\t%s" _line (nth _buffer _line))
       )
      )
      ((and (= (n _got_list) 2) (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (printf "%d\t%s" (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
      )
      ((and (= (n _got_list) 3) (string->number (cadr _got_list)) (string->number (caddr _got_list)))
       (when (and (nth _buffer (string->number (cadr _got_list))) (nth _buffer (string->number (caddr _got_list))))
        (for _line (string->number (cadr _got_list)) (string->number (caddr _got_list)) 1 
         (printf "%d\t%s" _line (nth _buffer _line))
        )
       )
      )
      (true)
     )      
    )
    
    ; Switch to DOS/Mac format
 
    ((eq? (car _got_list) "o")
     (set! _previous_buffer _buffer)
     (set! _buffer '())
     (for-each _string _previous_buffer
      (if (string/ends-with? _string "\r")
       (push! _buffer _string)
       (push! _buffer (sprintf "%s\r" _string))
      )
     )
    )
 
    ; Paste content of clipboard from last or from any line
    
    ((eq? (car _got_list) "p")
     (if (= (n _got_list) 1)
      (progn
       (set! _previous_buffer _buffer)
       (set! _buffer (append _buffer _clipboard))
      )
      (when (and (string->number (cadr _got_list)) (nth _buffer (string->number (cadr _got_list))))
       (set! _previous_buffer (snoc _buffer))
       (for-each _string (reverse _clipboard)
        (insert-nth! _buffer (string->number (cadr _got_list)) _string)
       )
      )
     )
    )
    
    ; Read a file and append content into buffer
    
    ((eq? (car _got_list) "r")
     (when (and (= (n _got_list) 2) (not (empty? (string/trim (cadr _got_list)))))
      (with (_chid (io/open (string/trim (cadr _got_list))))
       (if (not _chid)
        (print! "File read failed.")
        (progn
         (io/close _chid)
         (set! _previous_buffer _buffer)
         (set! _buffer (append _buffer (read-lines! (string/trim (cadr _got_list)))))
         (print! "File read successful.")
        )
       )
      )
     )
    )
    
    ; Search a string or substitute it with another one (support Lua patterns) in all or some or only one line
 
    ((eq? (car _got_list) "s")
     (cond
      ((= (n _got_list) 2) 
       (for _line 1 (n _buffer) 1
        (when (and (pcall string/match (nth _buffer _line) (cadr _got_list)) (string/match (nth _buffer _line) (cadr _got_list)))
         (printf "%d\t%s" _line (nth _buffer _line))
        )
       )
      )
      ((= (n _got_list) 3) 
       (set! _previous_buffer (snoc _buffer))
       (for _line 1 (n _buffer) 1
        (when (and (pcall string/match (nth _buffer _line) (cadr _got_list)) (string/match (nth _buffer _line) (cadr _got_list)))
         (setq! (nth _buffer _line) (string/gsub (nth _buffer _line) (cadr _got_list) (caddr _got_list)))
        )
       )  
      )
      ((= (n _got_list) 4)
       (when (and (string->number (cadddr _got_list)) (nth _buffer (string->number (cadddr _got_list))))
        (when (and (pcall string/match (nth _buffer (string->number (cadddr _got_list))) (cadr _got_list)) (string/match (nth _buffer (string->number (cadddr _got_list))) (cadr _got_list)))
         (set! _previous_buffer (snoc _buffer))
         (setq! (nth _buffer (string->number (cadddr _got_list))) (string/gsub (nth _buffer (string->number (cadddr _got_list))) (cadr _got_list) (caddr _got_list)))
        )
       )
      )
      ((= (n _got_list) 5) 
       (when (and (string->number (cadddr _got_list)) (string->number (last _got_list)))
        (when (and (nth _buffer (string->number (cadddr _got_list))) (nth _buffer (string->number (last _got_list))))
         (set! _previous_buffer (snoc _buffer))
         (for _line (string->number (cadddr _got_list)) (string->number (last _got_list)) 1
          (when (and (pcall string/match (nth _buffer _line) (cadr _got_list)) (string/match (nth _buffer _line) (cadr _got_list)))
           (setq! (nth _buffer _line) (string/gsub (nth _buffer _line) (cadr _got_list) (caddr _got_list)))
          )
         )
        )
       )
      )
      (true)
     )
    )
 
    ; Show version and license
 
    ((eq? (car _got_list) "v")
     (print! "led version 20180101\n\nCopyright (c) 2017, Pasquale Frega\n\nAll rights reserved.\n\nReleased under the semplified BSD license.")
    )
 
    ; Replace tab with # spaces (default 1, usefull on Amiga's)
 
    ((eq? (car _got_list) "t")
     (set! _previous_buffer _buffer)
     (set! _buffer '())
     (let [(_times (string->number (cadr _got_list))) (_spaces nil)]
      (when (not _times)
       (set! _times 1)
      )
      (set! _spaces (string/rep " " _times))
      (for-each _string _previous_buffer
       (push! _buffer (car (list (string/gsub _string "\t" _spaces))))
      )
     )
    )
 
    ; Undo<->Redo
 
    ((eq? (car _got_list) "u")
     (with (_exchanger _buffer)
      (set! _buffer _previous_buffer)
      (set! _previous_buffer _exchanger)
     )
    )
 
    ; Write out unconditionally
    
    ((eq? (car _got_list) "w")
     (if (write-lines! _file_name _buffer)
      (print! "File written.")
      (print! "Unable to write file.")
     )
    )
    
    ; Execute file or a shell command + file, execute a shell command, with r record it
    
    ((eq? (car _got_list) "x")
     (cond
      ((= (n _got_list) 1)
       (execute (string/trim (sprintf "%s %s" _shell_command _file_name)))
      )
      ((and (= (n _got_list) 2) (not (empty? (string/trim (cadr _got_list)))))
       (execute (string/trim (cadr _got_list)))
      )
      (true
       (set! _shell_command (string/trim (cadr _got_list)))
      )
     )
    )
    
    (true)
   )
  )
 )
)

; end